import signal
from contextlib import contextmanager

import psutil


def _session_terminator(signum, frame):
    print(f"Encountered Signal {signum}!")
    print("Terminating SSM session")
    for to_be_terminated in psutil.process_iter():
        if to_be_terminated.name() == "session-manager-plugin":
            to_be_terminated.terminate()
            print(">>> Terminated")


@contextmanager
def interrupt_handler():
    signal.signal(signal.SIGINT, _session_terminator)
    signal.signal(signal.SIGABRT, _session_terminator)
    signal.signal(signal.SIGTERM, _session_terminator)
    yield



import getpass
import platform
from datetime import datetime
from typing import Dict, Final, Optional

import boto3
from boto3.session import Session
from jdict import jdict, patch_module

from .user_data import fetch, store

patch_module("botocore.parsers")
patch_module("botocore.configloader")


_STS_CLIENT: Final = boto3.client("sts")


class Credentials(jdict):
    AccessKeyId: str
    SecretAccessKey: str
    SessionToken: str
    Expiration: str
    Region: str


def _make_file_name(profile: str) -> str:
    return f"aws-{profile}-credentials"


def _read_credentials(profile: str) -> Optional[Credentials]:
    try:
        credentials = fetch(_make_file_name(profile))
        expiration = datetime.fromisoformat(credentials.Expiration)
        return credentials if expiration > datetime.now(expiration.tzinfo) else None
    except FileNotFoundError:
        return None


def _get_config(profile: str) -> jdict:
    return Session(profile_name=profile)._session.get_scoped_config()


def _get_mfa_serial(profile: str) -> str:
    """
    Gets MFA serial ARN via client STS request
    """
    sts = jdict(Session(profile_name=profile).client("sts").get_caller_identity())
    return sts.Arn.replace("user", "mfa", 1)


def _store_credentials(profile: str, credentials: Credentials) -> None:
    credentials.Expiration = datetime.isoformat(
        credentials.Expiration
    )  # to make it json serializable
    store(_make_file_name(profile), credentials)


def _get_profile_credentials(profile: str, config: jdict) -> Credentials:
    token_code = input("Enter MFA Code: ")

    credentials = (
        Session(
            aws_access_key_id=config.aws_access_key_id,
            aws_secret_access_key=config.aws_secret_access_key,
            profile_name=profile,
            region_name=config.region,
        )
        .client("sts")
        .get_session_token(SerialNumber=_get_mfa_serial(profile), TokenCode=token_code)
        .Credentials
    )

    credentials.Region = config.region
    _store_credentials(profile, credentials)
    return credentials


def _get_role_credentials(profile: str, config: jdict) -> Credentials:
    source_profile = config.source_profile
    source_config = _get_config(source_profile)
    source_credentials = _read_credentials(source_profile) or _get_profile_credentials(
        source_profile, source_config
    )
    credentials = (
        _make_session(source_credentials)
        .client("sts")
        .assume_role(
            RoleArn=config.role_arn,
            RoleSessionName=f"{getpass.getuser()}@{platform.node()}-session",
        )
        .Credentials
    )
    credentials.Region = source_config.region
    _store_credentials(profile, credentials)
    return credentials


def _get_credentials(profile: str) -> Dict:
    config = _get_config(profile)
    return (
        _get_role_credentials(profile, config)
        if "role_arn" in config
        else _get_profile_credentials(profile, config)
    )


def _make_session(credentials) -> Session:
    return Session(
        aws_access_key_id=credentials.AccessKeyId,
        aws_secret_access_key=credentials.SecretAccessKey,
        aws_session_token=credentials.SessionToken,
        region_name=credentials.Region,
    )


def get_session(kwargs: Dict[str, str]) -> Session:
    profile = kwargs.get("profile", "default")
    return _make_session(_read_credentials(profile) or _get_credentials(profile))



from typing import Any, Dict, Generator, Iterator, Tuple

from boto3 import resource
from instances_map_abc.vm_instance_mapping import VmInstanceMappingBase
from instances_map_abc.vm_instance_proxy import VmInstanceProxy

from .ec2_instance_proxy import Ec2InstanceProxy, Ec2RemoteShellProxy


class Ec2AllInstancesData:
    def __init__(self) -> None:
        self._all_instances = resource("ec2").instances.all()
        self._instances_data = [
            (
                _instance.id,
                self._get_instance_name(_instance),
                _instance.state.Code,
                _instance.state.Name,
            )
            for _instance in self._all_instances
        ]

    def __iter__(self) -> Iterator:
        yield from self._instances_data

    def _get_instance_name(self, _instance: Dict[str, Any]) -> str:
        for tag in _instance.tags:
            if tag["Key"] == "Name":
                return tag["Value"]


class Ec2InstanceMapping(VmInstanceMappingBase[VmInstanceProxy]):
    def __init__(self, session) -> None:
        self._session = session
        self._client = self._session.client("ec2")

    def __getitem__(self, name: str) -> VmInstanceProxy:
        instance_id = self._get_instance_id(name)
        return self._get_instance(instance_id)

    def __iter__(self) -> Iterator:
        instances = (
            r["Instances"][0] for r in self._client.describe_instances()["Reservations"]
        )
        for instance in instances:
            yield self._get_instance(instance["InstanceId"])

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def keys(self) -> Generator[str, None, None]:
        for instance in self:
            yield instance.name

    def values(self) -> Generator[str, None, None]:
        yield from self

    def items(self) -> Generator[Tuple[str, str], None, None]:
        yield from zip(self.keys(), self.values())

    def _get_instance(self, instance_id: str) -> Ec2InstanceProxy:
        return Ec2InstanceProxy(instance_id, self._session, ec2_client=self._client)

    def _get_instance_id(self, instance_name: str) -> str:
        instance_details = self._client.describe_instances(
            Filters=[
                {
                    "Name": "tag:Name",  # as long as you are following the convention of putting Name in tags
                    "Values": [
                        instance_name,
                    ],
                },
            ],
        )
        return instance_details["Reservations"][0]["Instances"][0]["InstanceId"]


class Ec2RemoteShellMapping(Ec2InstanceMapping, VmInstanceMappingBase):
    def _get_instance(self, instance_id: str) -> Ec2RemoteShellProxy:
        return Ec2RemoteShellProxy(instance_id, self._session)



import time
from typing import Any, Iterable, Optional, Tuple, Union

import botocore
from boto3 import resource, setup_default_session
from instances_map_abc.vm_instance_proxy import VmState


class _Ec2StateProxy(VmState):
    pending = 0
    running = 16
    shutting_down = 32
    terminated = 48
    stopping = 64
    stopped = 80


class Ec2InstanceProxy:
    def __init__(
        self,
        instance_id: str,
        session,
        ec2_client: Optional[botocore.client.BaseClient] = None,
        **kwargs: str,
    ) -> None:
        self._instance_id = instance_id
        self._ec2_client = ec2_client or session.client("ec2")
        # ---
        setup_default_session(profile_name=session.profile_name)
        self._instance = resource("ec2").Instance(instance_id)

    def start(self, wait: bool = True) -> None:
        """
        Start the vm

        :return: None
        """
        self._ec2_client.start_instances(
            InstanceIds=[self._instance_id],
        )
        wait and self._instance.wait_until_running()  # python short circuiting

    def stop(self, wait: bool = True) -> None:
        """
        Stop the vm

        :return: None
        """
        self._ec2_client.stop_instances(
            InstanceIds=[self._instance_id],
        )
        wait and self._instance.wait_until_stopped()

    @property
    def state(self) -> _Ec2StateProxy:
        return _Ec2StateProxy[self._instance.state["Name"]]

    @property
    def id(self) -> str:
        return self._instance_id

    @property
    def name(self):
        for tag in self._instance.tags:
            if tag["Key"] == "Name":
                return tag["Value"]


class Ec2RemoteShellProxy(Ec2InstanceProxy):
    def __init__(self, instance_id: str, session) -> None:
        super().__init__(instance_id, session)
        self._session = session
        self._ssm_client = self._session.client("ssm")

    def execute(
        self, *commands: Union[str, Iterable], **kwargs: str
    ) -> Union[Tuple[Any, ...], Tuple[str, str]]:
        result = self._ssm_client.send_command(
            InstanceIds=[self._instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": ["source /etc/bashrc", *commands]},
        )

        command_id = result["Command"]["CommandId"]
        # see https://stackoverflow.com/questions/50067035/retrieving-command-invocation-in-aws-ssm
        time.sleep(2)
        if not kwargs.get("wait", True):
            return self._ssm_client, command_id, self._instance_id
        waiter = self._ssm_client.get_waiter("command_executed")
        try:
            waiter.wait(
                CommandId=command_id,
                InstanceId=self._instance_id,
                WaiterConfig={
                    "Delay": kwargs.get("delay", 5),
                    "MaxAttempts": kwargs.get("attempts", 20),
                },
            )
        finally:
            result = self._ssm_client.get_command_invocation(
                CommandId=command_id,
                InstanceId=self._instance_id,
                PluginName="aws:RunShellScript",
            )
            print(result.StandardOutputContent)
            print(result.StandardErrorContent)
        return result.StandardOutputContent, result.StandardErrorContent

    @property
    def session(self):
        return self._session



"""
User Data Access Utility
Adopted from https://gist.github.com/jslay88/1fd8a8ba1d05ff2a4810520785a67891
"""
import json
import pathlib
import sys
from typing import Dict, Final

from jdict import jdict, set_json_decoder

_HOME: Final[pathlib.Path] = pathlib.Path.home()

_SYSTEM_USER_DIR: Final[Dict[str, str]] = {
    "win32": "AppData/Roaming",
    "linux": ".local/share",
    "darwin": "Library/Application Support",
}

_SYSTEM_SSM_PATH: Final[pathlib.Path] = _HOME / _SYSTEM_USER_DIR[sys.platform] / "ssm"

set_json_decoder(json)

_SYSTEM_SSM_PATH.mkdir(parents=True, exist_ok=True)


def _get_path(name: str) -> pathlib.Path:
    return _SYSTEM_SSM_PATH / f"{name}.json"


def fetch(name: str) -> jdict:
    with _get_path(name).open("r") as file:
        return json.load(file)


def store(name: str, data: jdict) -> None:
    with _get_path(name).open("w") as file:
        json.dump(data, file)


def remove(name: str) -> None:
    _get_path(name).unlink(missing_ok=True)
